"""General Constants"""
from enum import Enum

URL_LOGIN = "https://www.trading212.com/en/login"
CLASS_COOKIES_NOTICE_BUTTON = "CookiesNotice_button__q5YaL"
XPATH_LOGIN_BUTTON = '//*[@id="root"]/div/div/div/div/div/div/div[1]/div/div[2]/div/div/div/div[1]/div/div[1]/div[2]/div[4]/div/div'
CLASS_EQUITY_ICON = "equity-icon"
CLASS_CFD_ICON = "cfd-icon"
PYTRADING212_VERSION = "0.2.8"
XPATH_EMAIL = '//*[@id="root"]/div/div/div/div/div/div/div[1]/div/div[2]/div/div/div/div[1]/div/div[1]/div[2]/div[1]/div/div[1]/div[2]/input'
XPATH_PASSWORD = '//*[@id="root"]/div/div/div/div/div/div/div[1]/div/div[2]/div/div/div/div[1]/div/div[1]/div[2]/div[2]/div/div/div[1]/div[2]/input'
SELECTOR_MENU_BUTTON = ("#root > div > div > div > div > div > div > div:nth-child(1) > div > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div > div > "
                        "div:nth-child(1) > div > div > div:nth-child(2) > div > div:nth-child(5) > div > span > div")
SELECTOR_LIVE_ACCOUNT = ("#root > div > div > div > div > div > div > div:nth-child(1) > div > "
                         "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div > div > "
                         "div:nth-child(1) > div > div > div:nth-child(3) > div:nth-child(1) > div > div > div")
SELECTOR_LIVE_EQUITY = ("#root > div > div > div > div > div > div > div:nth-child(1) > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-105ug2t > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-105ug2t > div:nth-child(3) > "
                        "div > div.css-175oi2r.r-12vffkv > div > div > "
                        "div.css-175oi2r.r-150rngu.r-eqz5dr.r-16y2uox.r-11yh6sk.r-1rnoaur.r-1sncvnh > div > "
                        "div:nth-child(2) > div:nth-child(1) > div:nth-child(1) > div.css-175oi2r")
SELECTOR_DEMO_EQUITY = ("#root > div > div > div > div > div > div > div:nth-child(1) > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-105ug2t > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-105ug2t > div:nth-child(3) > "
                        "div > div.css-175oi2r.r-12vffkv > div > div > "
                        "div.css-175oi2r.r-150rngu.r-eqz5dr.r-16y2uox.r-11yh6sk.r-1rnoaur.r-1sncvnh > div > "
                        "div:nth-child(2) > div:nth-child(1) > div:nth-child(1)")
SELECTOR_MENU_LIST = ("#root > div > div > div > div > div > div > div:nth-child(1) > div > "
                      "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div > div > "
                      "div:nth-child(2) > div > div > div:nth-child(2) > div > div > div > div > div > "
                      "div.css-175oi2r.r-13awgt0 > div > "
                      "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div:nth-child(2) > "
                      "div > div > div > div.css-175oi2r.r-13awgt0 > div > div:nth-child(1)")
SELECTOR_SWITCH_DEMO = ("#root > div > div > div > div > div > div > div:nth-child(1) > div > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div > div > "
                        "div:nth-child(2) > div > div > div:nth-child(2) > div > div > div > div > div > "
                        "div.css-175oi2r.r-13awgt0 > div > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div:nth-child(2) > "
                        "div > div > div > div.css-175oi2r.r-13awgt0 > div > div:nth-child(1) > div > div > div > div "
                        "> div.css-175oi2r.r-13awgt0 > div > "
                        "div.css-175oi2r.r-1p0dtai.r-1d2f490.r-u8s1d.r-zchlnj.r-ipm5af.r-12vffkv > div:nth-child(2) > "
                        "div > div.css-175oi2r.r-13awgt0.r-1udh08x > div > div.css-175oi2r.r-13awgt0 > div > "
                        "div:nth-child(2) > div.css-175oi2r.r-150rngu.r-eqz5dr.r-16y2uox.r-1wbh5a2.r-11yh6sk.r"
                        "-1rnoaur.r-1sncvnh > div > div:nth-child(3) > "
                        "div.css-175oi2r.r-1otgn73.r-1loqt21.r-1awozwy.r-1777fci.r-utggzx")

ONE_WEEK = "ONE_WEEK"


class Mode(Enum):
    """Mode Type"""
    DEMO = "demo",
    LIVE = "live"


class Trading(Enum):
    """Trading Type"""
    CFD = "CFD",
    EQUITY = "EQUITY",


class Period(Enum):
    """Period of Portfolio Performance"""
    LAST_DAY = "LAST_DAY",
    LAST_WEEK = "LAST_WEEK",
    LAST_MONTH = "LAST_MONTH",
    LAST_THREE_MONTHS = "LAST_THREE_MONTHS",
    LAST_YEAR = "LAST_YEAR",
    ALL = "ALL",


class OrderType(Enum):
    """Order Type"""
    LIMIT = "LIMIT"
    STOP = "STOP"
    MARKET = "MARKET"
    STOP_LIMIT = "STOP_LIMIT"


class TimeValidity(Enum):
    """Time Validity for Limit/Stop Orders"""
    DAY = "DAY",
    GOOD_TILL_CANCEL = "GOOD_TILL_CANCEL"



import datetime
import json
import typing

import requests


class CandleStick(object):
    """
        The candlestick object is a representaion of the values used in a candlestick
        chart.

        It consists of an opening timestamp opening price, closing price as well
        as a high and low price.

        The trading212 api also gives an int value which seems to be 0 all the
        time which I have not figured out what this is but it is included as
        "unknown"
    """

    opening: float
    closing: float
    high: float
    low: float
    timestamp: datetime.datetime
    unknown: int

    def __init__(self, data: typing.List) -> None:
        self.timestamp, self.high, self.low, self.opening, self.closing, self.unknown = data

    def __repr__(self) -> str:
        return (f"CandleStick(timestamp={self.timestamp}, high={self.high}, low={self.low},"
                f" opening={self.opening}, closing={self.closing}, unknown={self.unknown})")


class Company:
    """Company Wrapper"""

    charting_url: str = "/charting/v3/candles"

    def __init__(self, instrument_code: str, isin: str, language_code: str = 'en'):
        self.instrument_code = instrument_code
        self.isin = isin
        self.language_code = language_code

    def get_pricing_history(
            self,
            interval: str,
            headers: dict,
            trading_instance: 'Trading212'
    ) -> typing.List[CandleStick]:
        ticker_data: str = json.dumps(
            {
                "candles":
                    [
                        {
                            "ticker": self.instrument_code,
                            "useAskPrice": True,
                            "period": interval,
                            "size": 500
                        }
                    ]
            }
        )

        response = requests.put(
            f"{trading_instance.base_url}{self.charting_url}?ticker={self.instrument_code}&languageCode={self.language_code}",
            headers=headers,
            data=ticker_data
        )
        return [
            CandleStick(x) for x in json.loads(
                response.content.decode("utf-8")
            )[0]["response"]["candles"]
        ]


class CFDInstrument:
    def __init__(self, name: str, ticker: str, sell_price: float, buy_price: float):
        self.name = name
        self.ticker = ticker
        self.sell_price = sell_price
        self.buy_price = buy_price



"""API for Trading212 Platform"""

import json
import logging
import re
import time
from datetime import datetime
from time import strftime
from urllib.parse import urlencode

import requests
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from pytrading212 import constants, console
from pytrading212.instrument import CFDInstrument
from pytrading212.order import CFDOrder, EquityOrder
from pytrading212.position import Position


class Trading212:
    def __init__(
            self,
            email: str,
            password: str,
            driver: webdriver,
            mode: constants.Mode = constants.Mode.DEMO,
            trading: constants.Trading = constants.Trading.EQUITY,
    ):
        self.session = f"TRADING212_SESSION_{mode.name}"
        self.base_url = f"https://{mode.name.lower()}.trading212.com"

        console.log(f"Starting PyTrading212 v{constants.PYTRADING212_VERSION}\n"
                    f"Trading: [green]{trading.name}[/green] \n"
                    f"Mode: [green]{mode.name}[/green]")

        self.driver = driver
        self.driver.get(constants.URL_LOGIN)

        # Click Accept all cookies if it appears
        try:
            console.log("Closing 'Cookies Popup'")
            self.driver.find_element(By.CLASS_NAME, constants.CLASS_COOKIES_NOTICE_BUTTON).click()
        except NoSuchElementException:
            pass  # ignore

        console.log("Authenticating")

        WebDriverWait(self.driver, 120).until(expected_conditions.visibility_of_element_located((By.XPATH, constants.XPATH_EMAIL)))

        # Authenticate
        self.driver.find_element(By.XPATH, constants.XPATH_EMAIL).send_keys(email)
        self.driver.find_element(By.XPATH, constants.XPATH_PASSWORD).send_keys(password)

        # Click login button
        self.driver.find_element(By.XPATH, constants.XPATH_LOGIN_BUTTON).click()

        self.user_agent = self.driver.execute_script("return navigator.userAgent;")

        # Switch to get also DEMO token
        self.switch()
        # Get session cookie
        cookies = self.driver.get_cookies()
        if cookies is not None:
            for cookie in cookies:
                # Get appropriate cookie for this session, live or demo
                if self.session in cookie['name']:
                    self.cookie = f"{self.session}={cookie['value']};"
        else:
            raise Exception("Unable to get cookies, aborting.")

        # Necessary headers for requests
        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": self.user_agent,
            "Cookie": self.cookie,
        }

        self.companies = self.get_companies()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()

    def finish(self):
        console.log("Closing session.")
        self.driver.close()

    def switch(self):
        WebDriverWait(self.driver, 30).until(expected_conditions.visibility_of_element_located(
            (By.CSS_SELECTOR, constants.SELECTOR_MENU_BUTTON)))
        self.driver.find_element(By.CSS_SELECTOR, constants.SELECTOR_MENU_BUTTON).click()

        self.driver.execute_script("arguments[0].scroll(0, arguments[0].scrollHeight);",
                                   self.driver.find_element(By.CSS_SELECTOR, constants.SELECTOR_MENU_LIST))
        self.driver.find_element(By.CSS_SELECTOR, constants.SELECTOR_SWITCH_DEMO).click()
        WebDriverWait(self.driver, 30).until(expected_conditions.visibility_of_element_located(
            (By.CSS_SELECTOR, constants.SELECTOR_DEMO_EQUITY)))
        self.driver.find_element(By.CSS_SELECTOR, constants.SELECTOR_DEMO_EQUITY).click()
        # Wait some seconds to initialize cookies
        time.sleep(5)

    def get_funds(self):
        """Get your funds, free, available."""
        response = requests.get(
            f"{self.base_url}/rest/v2/customer/accounts/funds", headers=self.headers
        )
        return json.loads(response.content.decode("utf-8"))

    def last_hour_hotlist(self):
        """Trading 212 last hour hotlist"""
        response = requests.get(
            f"{self.base_url}/trading212.com/rest/positions-tracker/deltas/hourly/1"
        )
        return json.loads(response.content.decode("utf-8"))

    def get_orders(self, older_than: datetime, newer_than: datetime):
        """Get orders within a range of dates"""
        params = {
            'olderThan': strftime(older_than.isoformat()),
            'newerThan': strftime(newer_than.isoformat())
        }

        response = requests.get(
            f"{self.base_url}/rest/history/orders", headers=self.headers, params=urlencode(params)
        )
        return json.loads(response.content.decode("utf-8"))

    def get_transactions(self, older_than: datetime, newer_than: datetime):
        """Get transactions within a range of dates"""
        params = {
            'olderThan': strftime(older_than.isoformat()),
            'newerThan': strftime(newer_than.isoformat())
        }
        response = requests.get(
            f"{self.base_url}/rest/history/transactions", headers=self.headers, params=urlencode(params)
        )
        return json.loads(response.content.decode("utf-8"))

    def get_order_details(self, details_path):
        """Get Order Details"""
        response = requests.get(f"{self.base_url}/rest/history{details_path}", headers=self.headers)
        return json.loads(response.content.decode("utf-8"))

    def get_dividends(self, older_than: datetime, newer_than: datetime):
        """Get dividends within a range of dates"""
        params = {'olderThan': strftime(older_than.isoformat()),
                  'newerThan': strftime(newer_than.isoformat())
                  }

        response = requests.get(
            f"{self.base_url}/rest/history/dividends", headers=self.headers, params=urlencode(params)
        )
        return json.loads(response.content.decode("utf-8"))

    def get_fundamentals(self, ticker, language_code: str = "en"):
        """Get fundamentals of a company by its isin"""
        params = {'ticker': ticker,
                  'languageCode': language_code
                  }

        response = requests.get(f"{self.base_url}/rest/companies/v2/fundamentals",
                                params=params)
        return json.loads(response.content.decode("utf-8"))

    def get_portfolio_performance(self, time_period: constants.Period):
        """Get Portfolio Performance"""
        response = requests.get(
            url=f"{self.base_url}/rest/v2/portfolio?period={time_period}",
            headers=self.headers,
        )
        return json.loads(response.content.decode("utf-8"))

    def get_portfolio_composition(self):
        """Get Portfolio Composition"""
        # click portfolio section on right-sidepanel
        right_sidepanel_portfolio_class = 'portfolio-icon'
        condition = expected_conditions.visibility_of_element_located(
            (By.CLASS_NAME, right_sidepanel_portfolio_class)
        )
        WebDriverWait(self.driver, 30).until(condition)
        self.driver.find_element(By.CLASS_NAME, right_sidepanel_portfolio_class).click()

        positions = []
        try:
            # click on investments
            self.driver.find_element(By.CLASS_NAME, 'investment-tab').click()
            for item in self.driver.find_elements(By.CLASS_NAME, "investment-item"):
                ticker = item.get_attribute("data-qa-item")
                value = item.find_element(By.CLASS_NAME, "total-value").text
                quantity = item.find_element(By.CLASS_NAME, "quantity").text
                total_return = item.find_element(By.CLASS_NAME, "return").text
                position = Position(ticker, value, quantity, total_return)
                positions.append(position.__dict__)
        except Exception as e:
            logging.error(e)  # portfolio is empty
        return positions

    def get_companies(self):
        """Get Ticker of all Trading212 tradable companies"""
        response = requests.get(
            url=f"{self.base_url}/rest/companies",
            headers=self.headers
        )
        return json.loads(response.content.decode("utf-8"))

    def search(self, query):
        """Search a company"""
        found = []
        companies = self.get_companies()
        for company in companies:
            if query in company["ticker"]:
                found.append(company["ticker"])

        response = requests.post(
            f"{self.base_url}/charting/prices?withFakes=false",
            headers=self.headers,
            data=found.__str__().replace("'", '"'),
        )  # ' with " for Trading212 compatibility
        return json.loads(response.content.decode("utf-8"))


class Equity(Trading212):
    """Trading 212 Equity"""

    def __init__(self, email: str, password: str, driver: webdriver, mode: constants.Mode = constants.Mode.DEMO):
        super().__init__(email, password, driver, mode, constants.Trading.EQUITY)

    def review_order(self, order: EquityOrder):
        """Preview of the order, with added costs and other useful data"""
        # Check if it is a 'value' order or a 'quantity' order
        if hasattr(order, 'value'):
            url = f"{self.base_url}/rest/v1/equity/value-order/review"
        else:
            url = f"{self.base_url}/rest/public/added-costs"
        response = requests.post(
            url=url,
            headers=self.headers,
            data=order.to_json(),
        )
        return json.loads(response.content.decode("utf-8"))

    def execute_order(self, order: EquityOrder):
        """Execute equity order"""
        # Check if it is a 'value' order or a 'quantity' order
        if hasattr(order, 'value'):
            url = f"{self.base_url}/rest/v1/equity/value-order"
        else:
            url = f"{self.base_url}/rest/public/v2/equity/order"
        response = requests.post(
            url=url,
            headers=self.headers,
            data=order.to_json(),
        )
        return json.loads(response.content.decode("utf-8"))

    def cancel_order(self, order_id):
        """Cancel a pending order"""
        response = requests.delete(
            url=f"{self.base_url}/rest/public/v2/equity/order/{order_id}",
            headers=self.headers,
        )
        return json.loads(response.content.decode("utf-8"))

    def check_order(self, equity_order: EquityOrder) -> [bool, str]:
        """Check if Order is valid."""
        is_valid_ticker = False, f"Instrument Code {equity_order.instrument_code} is not a valid Trading212 Ticker"
        for company in self.get_companies():
            if company['ticker'] == equity_order.instrument_code:
                is_valid_ticker = True, f"Instrument Code {equity_order.instrument_code} is valid Trading212 Ticker"

        return is_valid_ticker

    def min_max_sell_buy(self, instrument_code: str):
        params = {'instrumentCode': instrument_code}
        response = requests.get(
            f"{self.base_url}/rest/v1/equity/value-order/min-max",
            headers=self.headers,
            params=params
        )
        return json.loads(response.content.decode("utf-8"))


class CFD(Trading212):
    """ Trading 212 CFD """

    def __init__(self, email: str, password: str, driver: webdriver, mode: constants.Mode = constants.Mode.DEMO):
        super().__init__(email, password, driver, mode, constants.Trading.CFD)

    def execute_order(self, order: CFDOrder):
        """Execute CFD order"""

        # Check if it is Limit Stop Order
        if hasattr(order, 'is_limit_stop') and order.is_limit_stop == True:
            url = f"{self.base_url}/rest/v2/pending-orders/entry-dep-limit-stop/{order.instrument_code}"
        # Check if it is OCO Order
        elif hasattr(order, 'is_oco') and order.is_oco == True:
            url = f"{self.base_url}/rest/v2/pending-orders/entry-oco/{order.instrument_code}"
        # Market Order
        else:
            url = f"{self.base_url}/rest/v2/trading/open-positions"

        response = requests.post(
            url=url,
            headers=self.headers,
            data=order.to_json(),
        )
        return json.loads(response.content.decode("utf-8"))

    def trading_additional_info(self, order: CFDOrder):
        """Get additional trading info before executing order."""
        params = {'instrumentCode': order.instrument_code,
                  'quantity': order.quantity,
                  'positionId': 'null'}
        response = requests.get(
            f"{self.base_url}/rest/v1/tradingAdditionalInfo",
            headers=self.headers,
            params=params
        )
        return json.loads(response.content.decode("utf-8"))

    def close_position(self, position_id):
        """Close an open position."""
        response = requests.delete(
            url=f"{self.base_url}/rest/v2/trading/open-positions/close/{position_id}",
            headers=self.headers,
        )
        return json.loads(response.content.decode("utf-8"))

    def get_current_price(self, instrument_code):
        """Workaround to get the current price of a CFD."""
        # Simulate an order with target price 0, T212 will respond with a business exception so we can get the
        # current price
        cfd_order = CFDOrder(instrument_code=instrument_code,
                             target_price=0.0,
                             quantity=0.1)
        # Return only the current price
        return float(self.execute_order(cfd_order)['context']['current'])



class Position:
    def __init__(self, instrument_code, value, quantity, total_return):
        self.instrument_code = instrument_code
        self.value = value
        self.quantity = quantity
        self.total_return = total_return



from rich.console import Console

_console = Console()


def log(msg: str) -> None:
    """Takes a string and logs it to the console.

    Args:
        msg (str): The message to log.
    """
    _console.log(msg)



from .order import (
    EquityOrder,
    MarketOrder,
    LimitOrder,
    StopOrder,
    StopLimitOrder,
    ValueOrder,
    CFDMarketOrder,
    CFDLimitStopOrder,
    CFDOCOOrder
)
from .trading212 import Equity, CFD
from .constants import Mode, Trading, Period, OrderType, TimeValidity, ONE_WEEK
from .position import Position
from .instrument import CFDInstrument, Company



from pytrading212 import constants, utils


class Order:
    """Base Order"""

    def to_json(self):
        out = self.__dict__  # Convert order in dictionary
        # Replace ' with " for Trading212 compatibility
        return dict((utils.to_camel_case(key), value) for (key, value) in out.items()).__str__() \
            .replace("'", '"')


class EquityOrder(Order):
    """Base Equity Oder"""

    def __init__(self,
                 instrument_code: str,
                 order_type: constants.OrderType,
                 **kwargs):
        self.instrument_code = instrument_code
        self.order_type = order_type.name

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not (hasattr(self, 'quantity') or hasattr(self, 'value')):
            raise Exception("'value' or 'quantity' parameter must be be provided.")

        if hasattr(self, 'quantity') and hasattr(self, 'value'):
            raise Exception("'value' or 'quantity' both provided, only one is allowed.")

    def is_value_order(self):
        return hasattr(self, 'value')

    def is_quantity_order(self):
        return hasattr(self, 'quantity')


class MarketOrder(EquityOrder):
    """Market Order Wrapper."""

    def __init__(self, instrument_code: str, quantity: float):
        super().__init__(instrument_code=instrument_code,
                         order_type=constants.OrderType.MARKET,
                         quantity=quantity)


class LimitOrder(EquityOrder):
    """Limit Order Wrapper."""

    def __init__(
            self,
            instrument_code: str,
            quantity: float,
            limit_price: float,
            time_validity: constants.TimeValidity,
    ):
        super().__init__(instrument_code=instrument_code,
                         order_type=constants.OrderType.LIMIT,
                         quantity=quantity,
                         limit_price=limit_price,
                         time_validity=time_validity.name)


class StopOrder(EquityOrder):
    """Stop Order Wrapper."""

    def __init__(
            self,
            instrument_code: str,
            quantity: float,
            stop_price: float,
            time_validity: constants.TimeValidity,
    ):
        super().__init__(instrument_code=instrument_code, order_type=constants.OrderType.STOP,
                         quantity=quantity,
                         stop_price=stop_price,
                         time_validity=time_validity.name
                         )


class StopLimitOrder(EquityOrder):
    """Stop/Limit Order Wrapper."""

    def __init__(
            self,
            instrument_code: str,
            quantity: float,
            limit_price: float,
            stop_price: float,
            time_validity: constants.TimeValidity,
    ):
        super().__init__(instrument_code=instrument_code,
                         order_type=constants.OrderType.STOP_LIMIT, quantity=quantity, limit_price=limit_price,
                         stop_price=stop_price, time_validity=time_validity.name
                         )


class ValueOrder(EquityOrder):
    """Value Order Wrapper."""

    def __init__(self, instrument_code: str, value: float):
        super().__init__(instrument_code=instrument_code,
                         order_type=constants.OrderType.MARKET,
                         value=value)


class CFDOrder(Order):
    """Base CFD Oder"""

    def __init__(self, instrument_code: str, **kwargs):
        self.notify = "NONE"
        self.instrument_code = instrument_code

        for key, value in kwargs.items():
            setattr(self, key, value)


class CFDMarketOrder(CFDOrder):
    """CFD Market Order"""

    def __init__(self,
                 instrument_code: str,
                 quantity: float,
                 target_price: float,
                 **kwargs):
        super().__init__(instrument_code=instrument_code,
                         quantity=quantity,
                         target_price=target_price,
                         **kwargs)


class CFDLimitStopOrder(CFDOrder):
    """CFD Limit Stop Order (Pending Order)"""

    def __init__(self, instrument_code: str, quantity: float, target_price: float, **kwargs):
        super().__init__(instrument_code=instrument_code,
                         target_price=target_price,
                         quantity=quantity,
                         is_limit_stop=True,
                         **kwargs)

    def to_json(self):
        tmp = self
        delattr(tmp, 'instrument_code')  # Remove instrument code
        delattr(tmp, 'is_limit_stop')  # Remove is_limit_stop flag
        out = CFDOrder.to_json(tmp)
        return out


class CFDOCOOrder(CFDOrder):
    """CFD OCO Order"""

    class OCOSubOrder(Order):
        """CFD OCO Sub Order"""

        def __init__(self, price: float, quantity: float):
            self.price = price
            self.quantity = quantity

        def __repr__(self):
            return Order.to_json(self)

    def __init__(self, instrument_code: str, order1: OCOSubOrder, order2: OCOSubOrder):
        super().__init__(instrument_code=instrument_code,
                         order1=order1,
                         order2=order2,
                         is_oco=True)

    def to_json(self):
        tmp = self
        delattr(tmp, 'instrument_code')  # Remove instrument code
        delattr(tmp, 'is_oco')  # Remove is_oco flag
        out = CFDOrder.to_json(tmp)
        return out



def to_camel_case(text: str) -> str:
    """Convert a string to camel case.

    The first word in the text is converted to lowercase and
    the rest of the words are converted to title case, removing underscores.

    Args:
        text: The string to convert.

    Returns:
        The camel case string.
    """

    if "_" not in text:
        return text
    camel = "".join(
        word.capitalize() if i > 0 else word.lower()
        for i, word in enumerate(text.lstrip("_").split("_"))
    )
    return camel
